; Thrust-Kernel:
; This kernel was developed for a clone of the C-64 game THRUST
; Copyright 2000 Thomas Jentzsch
;
; Comments translated/added in 2001 (Sorry, not Andrew's quality ;-)
;
; *** Features: ***
; - Player 1:
;   - fixed
;   - updated every 2nd scanline
;   - y-position on _every_ scanline possible (VDELP1)
;
; - Player 0:
;   - multi-used
;   - updated every 2nd scanline
;   - x/y-position, color, vertical delay (or size), shape, fading changed for each copy
;   - some free cycles left (could be used for multi-coloring)
;
; - Ball & Missile 1: (Missile 0 not used)
;   - fixed position, variable width
;
; - Playfield:
;   - asymmetrical, not reflected
;   - full 160-pixel visible width
;   - max. height:~2048 pixel
;   - max. width:   272 pixel (increaseable, but needs some memory)
;   - striped, updated every 4th scanline
;   - (soft-)scrollable in both directions
;
; *** Limitations: ***
; - Player 1:
;   - some disturbations, when moved very far left (doesn't happen in this game)
;
; - Player 0:
;   - can start only every 4th scanline (realtive to PF), this can be improved
;      to every 2nd scanline with one extra 0 byte before shape-definition
;   - height must be even
;   - cannot end when playfield-block changes, this conflict can be evaded
;      with two extra 0 bytes after shape-definition
;   - data must be in one page
;
; - Ball & Missile 1:
;   - 2 pixel height
;
; - Playfield:
;   - each PF-line can only once be changed from on to off and back to on
;   - left and right border are defined in blocks of lines, only one new
;      block per PF-line possible
;      with clever block-arrangement this must not be a limitation


;===============================================================================
; A S S E M B L E R - S W I T C H E S
;===============================================================================

EXACTTIMING     = 0             ; if enabled, allows to remove WSYNC if neccessary


;===============================================================================
; C O N S T A N T S
;===============================================================================

MAXSPRITES  = 6                 ; max. number of visible player0
PFH         = 188               ; number of playfield-lines - 2 (4*n) (was 176)
PFW         = 68                ; width of playfield
SCW         = 40                ; width of screen


;===============================================================================
; Z P - V A R I A B L E S
;===============================================================================
    SEG.U   variables
    ORG     $80

tmpVar      .byte               ;                       ?
tmpVar2     .byte               ;                       ?
LineDiv2    = tmpVar            ; 0-PFH/2               8 kernel
PFgateLine  = tmpVar2           ; 0-PFH/2               8 kernel

;*** playfield variables ***
XOfsDiv4    .byte               ; 0..(PFW+SCW-1)        8 kernel
YOfsDiv4    .byte               ;                       8 kernel (for gate only)
PFoff       .byte               ; 0..(PFW+SCW*2-2)      7 kernel
PFon        .byte               ; 0..(PFW+SCW*2-2)      7 kernel
PFoffIdx    .byte               ;                       4 kernel
PFoffNxt    .byte               ;                       8 kernel
PFoffDiff   .byte               ; -2..+2                3 kernel
PFonIdx     .byte               ;                       4 kernel
PFonNxt     .byte               ;                       8 kernel
PFonDiff    .byte               ; -2..+2                3 kernel
PFgate      .byte               ;                       8 kernel

;*** ship variables ***
P1shape     .word               ;                     2*8 kernel
P1switch    .byte               ;                       8 kernel
P1end       .byte               ;                       8 kernel
P1move      .byte               ;                       8 kernel (start only)

;*** multi-used player variables ***
P0shape     .word               ;                     2*8 kernel
P0count     = P1move            ; 0..7/8, actual sprite 8 kernel

;----------------------------------------------
; lists for multi-used player 0:
; - y-start                         (x+1)
; - y-end                           (x+1)
; - shapepointer                    (x)
; - x-positioning                   (x+1)
; - color & vertical delay (or size)(x)
; - fading (only parts visible)     (x)
startList   .ds MAXSPRITES+1    ;                     x*8 kernel
endList     .ds MAXSPRITES      ; always odd!         x*8 kernel (last value in shapeList[0])
shapeList   .ds MAXSPRITES      ;                     x*8 kernel (hi-ptr constant!)
moveList    .ds MAXSPRITES+1    ;                     x*8 kernel
colorList   .ds MAXSPRITES      ;                     x*7+1 kernel
fadeList    .ds MAXSPRITES      ;                     x*8 kernel

P0start = startList
P0end   = endList
P0fade  = fadeList

BLline      .byte               ;                       8 kernel
M1line      .byte               ;                       8 kernel


;===============================================================================
; R O M - C O D E (BANK 0)
;===============================================================================
  SEG       Bank0
  ORG       $1000

;***** draw the last lines of the cave: *****
exitKernel:                 ;60
    iny
    sta     WSYNC
    sty     PF0             ; 3             y=0!
    sty     PF1             ; 3
    sty     PF2             ; 3
    sty     ENABL           ; 3
    sty     ENAM1           ; 3
    sty     REFP1           ; 3             enable PF-reflection for next frame
    ldx     #$ff            ; 2             restore stackpointer
    txs                     ; 2
    jmp     BankSwitch2     ;10 = 32        switch back to bank 2


    ALIGN 256
Kernel SUBROUTINE

; to LINE III, 2. calc PF-indices:
; special routine for the gates
.calcPF:                    ;11
    cpy     PFgateLine      ; 3
    beq     .gateStart      ; 2
    lda     PFonDiff        ; 3
    clc                     ; 2
    adc     PFon            ; 3
    sta     PFon            ; 3
    lda     PFoffDiff       ; 3
    clc                     ; 2
    bcc     .noGate         ; 3 = 35

.gateStart:                 ;   = 17
  IF EXACTTIMING
    lda.w   PFgate          ; 4
  ELSE
    lda     PFgate          ; 3
  ENDIF
    bpl     .gateOpening    ; 2
    bmi     .gateClosing    ; 2
.gateOpening:
    eor     #$ff            ; 2
.gateClosing:               ;   =  9/9

    cmp     #-GATEW         ; 2             completely opened?
    bcs     .gateNotOpen    ; 2
    lda     #1              ; 2
    bcc     .gateOpen       ; 2
.gateNotOpen:
    adc     #GATEW          ; 2             C=1!
    clc                     ; 2
.gateOpen:                  ;   =  9/9

.noGate:
  IF EXACTTIMING
    nop                     ; 2
    nop                     ; 2
    nop                     ; 2
    nop                     ; 2 =  8        waste some time
  ENDIF
    adc     PFoff           ; 3             C=0!
    sta     PFoff           ; 3
    jmp     .contPF         ; 3 =  9        -> 52 (bxx not possible, no constant flag)

.nxtOff:                    ; 6
    dec     PFoffIdx        ; 5
    ldx     PFoffIdx        ; 3
    lda     XOfsDiv4        ; 3
    adc     offInitTab,x    ; 4             C=1! (-> -1)
    sta     PFoff           ; 3

    adc     offWidthTab,x   ; 4
    sta     PFon            ; 3

    tya                     ; 2
    sec                     ; 2             optimizable!
    sbc     offNxtTab,x     ; 4             C = 1!
    sta     PFoffNxt        ; 3             C = ? (possible underflow)
    lda     offDiffTab,x    ; 4
    sta     PFoffDiff       ; 3
    jmp     .contPF         ; 3 = 52        (???) (bxx not possible, no constant flag)

;-----------------------------------------------------------
; The kernel consists out of four repated lines,
;  timing has to be quite accurate

;*************** LINE 0 ****************
; 1.-4. s.b.                        = 30
; 5. draw player 0 or wait          = 23
; 6. linecounter (part 2)           =  5
; 7. WSYNC                          =  3 (optimizable)
; SUM (<=76)                       -> 61 (15+3 free)
;***************************************
; 5. draw player 0 or wait:
.loopKernel:                ;10
    lda     (P0shape),y     ; 5
    and     P0fade          ; 3
    cpy     P0start         ; 3
    bcc     .contP0_0       ; 2
    lda     #0              ; 2
.contP0_0:                  ;   = 24/25

; 6. linecounter (part 2):
    dey                     ; 2
    sty     LineDiv2        ; 3 =  5

    sta     WSYNC           ; 3 =  3

;************** LINE III ***************
; 1. set GRP0                       =  3
; 2. calc PF-indices                = 52
; 3. draw player 1 or wait (part 1) = 15
; 4. WSYNC                          =  3 (optimizable)
; SUM (<=76)                       -> 73 (3+3 free)
;***************************************
; 1. set GRP0:
    sta     GRP0            ; 3 =  3        (if possible <= 22, ok)

; 2. calc PF-indices:
    cpy     PFoffNxt        ; 3
    beq     .nxtOff         ; 2=  5/ 6
    cpy     PFonNxt         ; 3
    bne     .calcPF         ; 2= 10/11

.nxtOn:                     ;10
    dec     PFonIdx         ; 5
    ldx     PFonIdx         ; 3
    lda     XOfsDiv4        ; 3
    adc     onInitTab,x     ; 4             C=1! (-> -1)
    sta     PFon            ; 3

    adc     onWidthTab,x    ; 4
    sta     PFoff           ; 3

    tya                     ; 2
    sbc     onNxtTab,x      ; 4             C=1!
    sta     PFonNxt         ; 3             C=? (possible underflow)
    lda     onDiffTab,x     ; 4
  IF EXACTTIMING
    sta.w   PFonDiff        ; 3
  ELSE
    sta     PFonDiff        ; 4
  ENDIF
.contPF:                    ;   = 52/52/52/52

; 3. draw player 1 or wait (part 1):
; some rather complicated stuff -> constant timing: 18 cycles
;  today I'd use a simplier solution :)
    cpy     P1switch        ; 3
    beq     .switchP1_2     ; 2
  IF EXACTTIMING
    bpl     .waitP1_2a      ; 2
  ELSE
    bpl     .waitP1_2       ; 2
  ENDIF
    lda     (P1shape),y     ; 5
    bcc     .contP1_2       ; 3

  IF EXACTTIMING
.waitP1_2a:
    nop                     ; 2
    bpl     .waitP1_2       ; 3
  ENDIF

.switchP1_2:
    lda     P1end           ; 3
  IF EXACTTIMING
    sta.w   P1switch        ; 4
  ELSE
    sta     P1switch        ; 3
  ENDIF
.waitP1_2:
    lda     #0              ; 2
.contP1_2:                  ;   = 15

    sta     WSYNC           ; 3 =  3

;************** LINE II ****************
; 1. draw player 1 or wait (part 2) =  3
; 2. draw missiles                  = 12
; 3. draw player 0 or wait          = 13
; 4. load PF-indices                =  6
; 5. draw playfield (part 1)        =  6
; 6. WSYNC                          =  3 (optimizable)
; SUM (<=76)                       -> 43 (33+3 free)
;***************************************
; 1. draw player 1 or wait (part 2):
    sta     GRP1            ; 3 =  3        (if possible <= 22, ok!)

; 2. draw missiles:
; got this trick from Combat (txs happend below!)
    cpy     BLline          ; 3
    php                     ; 3             ENABL (if possible <= 22, ok)
    cpy     M1line          ; 3
    php                     ; 3 = 12        ENAM1 (if possible <= 22, ok!)

enterKernel:                ;
; this is the point where we enter the kernel (y = PFH/2, everything is set up)
; 3. draw player 0 or wait:
    lda     (P0shape),y     ; 5
    and     P0fade          ; 3
    cpy     P0start         ; 3             not neccessary to test P0end here!
    bcc     .contP0_2       ; 2
    lda     #0              ; 2
.contP0_2:                  ;   = 14/15

; 4. load PF-indices:
    ldx     PFoff           ; 3
.contLine2:
    ldy     PFon            ; 3 =  6

; 5. draw playfield (part 1):
; x = PFoff, y = PFon, a = GRP0
; four different source codes for four possible PF setups
    cpx     #PFW+SCW/2      ; 2
    bcs     .PFoffXonR      ; 2/4
    cpy     #PFW+SCW/2      ; 2 =  6

    sta     WSYNC           ; 3 =  3

;*************** LINE I ****************
; 1. set GRP0 (part 2)              =  3
; 2. draw playfield (part 2)        = 49
; 3. draw missiles (part 1)         =  2
; 4. linecounter (part 1)           =  7
; 5. draw player 1 or wait (part 1) = 15
; SUM (<=76)                       -> 76 (0 free)
;***************************************
; 1. set GRP0 (part 2):
    sta     GRP0            ; 3 =  3        (if possible <= 22, ok)

; 2. draw playfield (part 2):
    lda     PF0offL,x       ; 4
    bcs     .PFoffLonR      ; 2

.PFoffLonL:                 ; 2| 9          (x < 80 && y < 80)
; both borders in the left part:
    ora     PF0onL,y        ; 4
    sta     PF0             ; 3|16          (z <= 22)
    lda     PF1offL,x       ; 4
    ora     PF1onL,y        ; 4
    sta     PF1             ; 3|27          (z <= 27)
    lda     PF2offL,x       ; 4
    ora     PF2onL,y        ; 4
    sta     PF2             ; 3|38          (z <= 38)
; the right part is completely filled:
    lda     #$ff            ; 2
    sta     PF0             ; 3|43          (28 <= z <= 49)
    sta     PF1             ; 3|46          (38 <= z <= 54)
    jmp     .PFcont         ; 3             (bne=4!)

.PFoffLonR:                 ; 3|10          (x < 80 && y >= 80)
; the left part is filled only with left border:
    sta     PF0             ; 3|13
    lda     PF1offL,x       ; 4
    sta     PF1             ; 3|20
    lda     PF2offL,x       ; 4
    sta     PF2             ; 3|27
; the right part is filled only with right border:
    lda     PF0onR,y        ; 4
    sta     PF0             ; 3|34
    lda     PF1onR,y        ; 4
    sta     PF1             ; 3|41
    lda     PF2onR,y        ; 4
    bcs     .PFcont         ; 4             (bank crossed!)

.exitKernelJMP:
    jmp     exitKernel

.loopKernelJMP:
    jmp     .loopKernel     ; 3

    .ds     9+4, 0          ;               just some free bytes
;---------------------------------------------------------------
    ALIGN 256

.PFoffRonRR:                ;17             (x, y >= 100)
; the whole line is filled:
    ldx     #5              ; 2
.waitRR:
    dex
    bne     .waitRR         ;24
    beq     .PFcont         ; 6

.PFoffXonR:                 ; 6
    sta     WSYNC           ; 3

    sta     GRP0            ; 3 =  3        (if possible <= 22, ok)

    cpx     #PFW+SCW        ; 2             (80 <= x, y < 100 ???)
; the left part is completely filled:
    lda     #$ff            ; 2
    sta     PF0             ; 3|10
    sta     PF1             ; 3|13
    sta.w   PF2             ; 4|17
    bcs     .PFoffRonRR     ; 2
; both borders in the right part:
    lda     PF0offR,x       ; 4             (80 <= x < 100, 80 <= y < ?)
    ora     PF0onR,y        ; 4
    sta     PF0             ; 3|30
    lda     PF1offR,x       ; 4
    ora     PF1onR,y        ; 4
    sta     PF1             ; 3|41
    lda     PF2offR,x       ; 4
    ora     PF2onR,y        ; 4
.PFcont:
    sta     PF2             ; 3|52
.PFexit:                    ;   = 49/49/49/49

; 3. draw missiles (part 1):
    ldx     #ENABL          ; 2 = 2         preload x with for txs

; 4. linecounter (part 1):
    ldy     LineDiv2        ; 3
    dey                     ; 2
    bmi     .exitKernelJMP  ; 2|59/61 = 7

; 5. draw player 1 or wait (part 1):
    cpy     P1switch        ; 3
    beq     .switchP1_0     ; 2
    bpl     .waitP1_0a      ; 2
    lda     (P1shape),y     ; 5
    bcc     .contP1_0       ; 3

.waitP1_0a:
    nop                     ; 2
    bpl     .waitP1_0       ; 3

.switchP1_0:
    lda     P1end           ; 3
    sta.w   P1switch        ; 4
.waitP1_0:
    lda     #0              ; 2
.contP1_0:                  ;   = 15
; no WSYNC in this line!

;*************** LINE 0 ****************
; 1. draw player 1 or wait (part 2) =  3
; 2. draw missiles (part 2)         = 14
; 3. clear playfield                = 11
; 4. draw missiles (part 3)         =  2
; 5.-7. s.b.                        = 32/46
; SUM                              -> 62/76
;***************************************
; 1. draw player 1 (part 2):
    sta     GRP1            ; 3 =  3        (if possible <= 22, ok!)

; 2. draw missiles (part 2):
    txs                     ; 2
    cpy     BLline          ; 3
    php                     ; 3             ENABL (if possible <= 22, ok)
    cpy     M1line          ; 3
    php                     ; 3 = 14        ENAM1 (if possible <= 22, ok)

; 3. clear playfield:
    lda     #0              ; 2
    sta     PF0             ; 3             (52-76 <= x <= 22!)
    sta     PF1             ; 3             (21, must <= 27)
    sta     PF2             ; 3 = 11        (24, must <= 38)

; 4. draw missiles (part 3):
    txs                     ; 2 =  2

; 5. start, draw player 0 (part 1) or wait:
    cpy     P0end           ; 3
    bne     .loopKernelJMP  ; 2

;===============================================================================
; it's time to start (...s) a new player 0
; actually, this happens at the end of the previous player,
;  and the new player might start to be displayed some lines later

;************** LINE 0s ****************
; 1.-4. s.a.                        = 30
; 5. start player 0 (part 1)        = 43
; 6. clear GRP0                     =  3
; SUM (<=76)                       -> 76 (0 free)
;***************************************
; 5. start player 0 (part 1):
    inc     P0count         ; 5
    ldx     P0count         ; 3
    lda     fadeList,x      ; 4
    sta     P0fade          ; 3
    lda     shapeList,x     ; 4
    sta     P0shape         ; 3
    lda     startList,x     ; 4
    sta     P0start         ; 3
    lda     endList,x       ; 4
    sta     P0end           ; 3
    lda     #0              ; 2 = 43

; 6. clear GRP0:
    sta     GRP0            ; 3 =  3        (if possible <= 22, ok)
; no WSYNC in this line!

;************* LINE IIIs ***************
; 1. determine left/right           = 15      |16
; 2./6. position player 0           =  5/11-26| 7-27
; 3./2. calc PF-indices (part 1/2)  =  6      | 9
; 4./3. save sprite-collision       =  7      | 7
; 5./4. update linecounter          =  5      | 5
; 6./5 draw missiles (part 1/2)     =  6      | 3
; 7. WSYNC                          =  3 (must!)
; SUM (<=76)                       -> 47-68   |50-70 (8|6 free)
;***************************************
; two different cases for the x-positioning,
;  this gives some free cycles for other stuff

; 1. determine left/right:
    lda     moveList,x      ; 4
    sta     HMP0            ; 3             evt. verlagern
    and     #%00001111      ; 2
    beq     .contLeft0      ; 2
    sbc     #6              ; 2             C=1!
    bcs     .moveRight      ; 2= 15/16

; 2. position player 0:
    adc     #4              ; 2
    bmi     .contLeft1      ; 2
.waitLeft:
    sbc     #1              ; 2
    bcs     .waitLeft       ; 2
.contLeft0:
    clc                     ; 2
.contLeft1:
    lda     PFonDiff        ; 3

; write to RESPx before 22 = write at 22
    sta     RESP0           ; 3 = 5/11/16-26 (20==22/26/31..46!)

; 3. calc PF-indices (part 1/2):
    adc     PFon            ; 3
    sta     PFon            ; 3 =  6

; 4. save sprite-collision: (result not used!, ???)
    lda     CXP0FB          ; 3
    sta     moveList-1,x    ; 4 =  7        variables no longer used

; 5. update linecounter:
    dey                     ; 2
    sty     LineDiv2        ; 3 =  5

; 6. draw missiles (part 1/2):
    cpy     BLline          ; 3
    jmp     .moveEnd        ; 3 =  6
;-----------------------------------------------------------

.moveRight:                 ;16
.waitRight:
    sbc     #1              ; 2
    bcs     .waitRight      ; 2

    nop
    nop                     ;   =  4

; 2. calc PF-indices (part 1/2):
    lda     PFonDiff        ; 3
    adc     PFon            ; 3
    sta     PFon            ; 3 =  9

; 3. save sprite-collision: (unused!)
    lda     CXP0FB          ; 3
    sta     moveList-1,x    ; 4 =  7

; 4. update linecounter:
    dey                     ; 2
    sty     LineDiv2        ; 3 =  5

; 5. draw missiles (part 1/2):
    cpy     BLline          ; 3 =  3

; 6. position player 0:
; I thought WSYNC needs 5 cycles, so the latest possible
;  end for RESP0 would be 71, today I know better :)
    sta     RESP0           ; 3 =  7-27     (51-71!)
.moveEnd:

    sta     WSYNC           ; 3 =  3

;************** LINE IIs ***************
; 1. move player 0                  =  3
; 2. draw missiles (part 2/2)       =  9
; 3. draw player 1                  = 20
; 4. set player 0 color             = 10
; 5. calc PF-indices (part 2/2)     = 11
; 6. set GRP0 (=0) (part 1)         =  2
; 7. draw playfield (part 1)        = 12
; 8. WSYNC                          =  3 (optimizable)
; SUM (<=76)                       -> 70 (6+3 free)
;***************************************
; 1. move player 0:
    sta     HMOVE           ; 3 =  3

; 2. draw missiles:
    php                     ; 3             ENABL (possibly <= 22, ok)
    cpy     M1line          ; 3
    php                     ; 3 =  9        ENAM1 (possibly <= 22, ok)

; 3. draw player 1:
    cpy     P1switch        ; 3
    beq     .switchP1_2s    ; 2
    bpl     .waitP1_2s      ; 2
    lda     (P1shape),y     ; 5
    bcc     .contP1_2s      ; 3

.switchP1_2s:
    lda     P1end           ; 3
    sta     P1switch        ; 3
.waitP1_2s:
    lda     #0              ; 2
.contP1_2s:                 ;
    sta     GRP1            ; 3             (if possible <= 22; no: 27-30!)
    clc                     ; 2 = 17-20

; 4. set player 0 color:
    lda     colorList,x     ; 4
    sta     VDELP0          ; 3             LSB used for VDELP0 (pod only!)
    sta     COLUP0          ; 3 = 10        bit 1..7 used for COLUP0

; 5. calc PF-indices (part 2/2):
    lda     PFoffDiff       ; 3
    adc     PFoff           ; 3             C=0!
    sta     PFoff           ; 3
    tax                     ; 2 = 11

; 6. set GRP0 (=0) (part 1):
    lda     #0              ; 2 =  2

    jmp     .contLine2      ; 3+9


;===============================================================================
; R O M - T A B L E S (Bank 0)
;===============================================================================
    ALIGN 256

; some pattern tables for the playfield (data is overlapping!):
; 196 bytes: (92*$ff, 8*data, 72*$00, 8*data, 16*$ff)
PF2onR:                                             ;   0
    .ds     16,$ff
PF0onR:                                             ;  16
    .ds     04,$ff
PF2onL:                                             ;  20
    .ds     16,$ff
PF0onL:                                             ;  36
    .ds     PFW-SCW/2,$ff                           ; PFW-20
    .ds     16,$ff
    .byte   $ff,$fe,$fc,$f8
    .byte   $f0,$e0,$c0,$80

PF2offL:                                            ; 100
    .ds     16,$00
PF0offL:                                            ; 116
    .ds     PFW-SCW/2,$00                           ; PFW-20
    .ds     16,$00
    .byte   $00,$01,$03,$07
    .byte   $0f,$1f,$3f,$7f
    .ds     16,$ff
PF2offR = PF2offL - SCW/2                           ;  80
PF0offR = PF0offL - SCW/2                           ;  96

;---------------------------------------------------------------
    ALIGN 256

; 172 bytes (84*$ff, 8*data, 64*$00, 8*data, 8*$ff)
PF1onR:                                             ;   0
    .ds     20,$ff
PF1onL:                                             ;  20
    .ds     PFW-SCW/2,$ff                           ; PFW-20
    .ds     24,$ff
    .byte   $ff,$7F,$3f,$1f,$0f,$07,$03,$01

PF1offL:                                            ;  92
    .ds     PFW-SCW/2,$00                           ; PFW-20
    .ds     24,$00
    .byte   $00,$80,$c0,$e0,$f0,$f8,$fc,$fe
    .ds      8,$ff
PF1offR = PF1offL - SCW/2                           ;  72

;all shape data has to be aligned at 256*n + PFH/2!
; actually the ship data starts a bit earlier, because
; the ship is never displayed at the very top of the screen